from django.db import models
from edc_constants.choices import YES_NO
from edc_model.validators import datetime_not_future
from edc_utils import get_utcnow

from ..choices import DEVIATION_VIOLATION, REPORT_STATUS
from ..models import ActionsRequired, ProtocolViolations


class ProtocolIncidentModelMixin(models.Model):

    report_datetime = models.DateTimeField(
        verbose_name="Report Date and Time", default=get_utcnow
    )

    short_description = models.CharField(
        verbose_name="Provide a short description of this incident",
        max_length=35,
        null=True,
        blank=False,
        help_text=(
            (
                "Max 35 characters. Note: there is additional space below for "
                "a more detailed description"
            )
        ),
    )

    report_type = models.CharField(
        verbose_name="Type of incident", max_length=25, choices=DEVIATION_VIOLATION
    )

    safety_impact = models.CharField(
        verbose_name="Could this incident have an impact on safety of the participant?",
        max_length=25,
        choices=YES_NO,
    )

    safety_impact_details = models.TextField(
        verbose_name='If "Yes", provide details', null=True, blank=True
    )

    study_outcomes_impact = models.CharField(
        verbose_name="Could this incident have an impact on study outcomes?",
        max_length=25,
        choices=YES_NO,
    )

    study_outcomes_impact_details = models.TextField(
        verbose_name='If "Yes", provide details', null=True, blank=True
    )

    incident_datetime = models.DateTimeField(
        verbose_name="Date incident occurred",
        validators=[datetime_not_future],
        null=True,
        blank=False,
    )

    incident = models.ForeignKey(
        ProtocolViolations,
        verbose_name="Type of incident",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        related_name="+",
    )

    incident_other = models.CharField(
        null=True, blank=True, verbose_name="If other, please specify", max_length=75
    )

    incident_description = models.TextField(
        verbose_name="Describe the incident",
        null=True,
        blank=False,
        help_text="Describe in full. Explain how the incident happened, what occurred, etc.",
    )

    incident_reason = models.TextField(
        verbose_name="Explain the reason why the incident occurred",
        null=True,
        blank=False,
    )

    corrective_action_datetime = models.DateTimeField(
        verbose_name="Corrective action date and time",
        validators=[datetime_not_future],
        null=True,
        blank=True,
    )

    corrective_action = models.TextField(
        verbose_name="Corrective action taken", null=True, blank=True
    )

    preventative_action_datetime = models.DateTimeField(
        verbose_name="Preventative action date and time",
        validators=[datetime_not_future],
        null=True,
        blank=True,
    )

    preventative_action = models.TextField(
        verbose_name="Preventative action taken", null=True, blank=True
    )

    action_required = models.ForeignKey(
        ActionsRequired,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        related_name="+",
    )

    report_status = models.CharField(
        verbose_name="What is the status of this report?",
        max_length=25,
        choices=REPORT_STATUS,
    )

    reasons_withdrawn = models.TextField(null=True, blank=True)

    report_closed_datetime = models.DateTimeField(
        blank=True,
        null=True,
        validators=[datetime_not_future],
        verbose_name="Date and time report closed.",
    )

    class Meta:
        abstract = True
        verbose_name = "Protocol Incident"
        verbose_name_plural = "Protocol Incident"
        indexes = [
            models.Index(fields=["subject_identifier", "action_identifier", "site", "id"])
        ]



# Generated by Django 3.2.13 on 2022-09-27 01:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_protocol_incident", "0014_auto_20220927_0400"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="protocolviolations",
            name="edc_protoco_id_b2c86f_idx",
        ),
        migrations.AddIndex(
            model_name="protocolviolations",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_protoco_id_58f1f5_idx",
            ),
        ),
        migrations.AlterModelTable(
            name="actionsrequired",
            table=None,
        ),
        migrations.AlterModelTable(
            name="protocoldeviationviolation",
            table=None,
        ),
        migrations.AlterModelTable(
            name="protocolincident",
            table=None,
        ),
        migrations.AlterModelTable(
            name="protocolincidents",
            table=None,
        ),
        migrations.AlterModelTable(
            name="protocolviolations",
            table=None,
        ),
    ]



from edc_constants.constants import CLOSED
from edc_form_validators import INVALID_ERROR

from edc_protocol_incident.constants import WITHDRAWN


class IncidentFormvalidatorMixin:
    def validate_close_report(self):
        self.required_if(
            CLOSED,
            WITHDRAWN,
            field="report_status",
            field_required="corrective_action_datetime",
        )
        self.required_if(
            CLOSED, WITHDRAWN, field="report_status", field_required="corrective_action"
        )
        self.required_if(
            CLOSED,
            WITHDRAWN,
            field="report_status",
            field_required="preventative_action_datetime",
        )
        self.required_if(
            CLOSED, WITHDRAWN, field="report_status", field_required="preventative_action"
        )
        self.required_if(
            CLOSED, WITHDRAWN, field="report_status", field_required="action_required"
        )
        self.required_if(
            CLOSED, WITHDRAWN, field="report_status", field_required="report_closed_datetime"
        )
        self.required_if(WITHDRAWN, field="report_status", field_required="reasons_withdrawn")

        self.validate_date_not_before_incident("report_closed_datetime")

    def validate_date_not_before_incident(self, fld_name):
        if self.cleaned_data.get(fld_name) and self.cleaned_data.get("incident_datetime"):
            if self.cleaned_data.get(fld_name) < self.cleaned_data.get("incident_datetime"):
                self.raise_validation_error(
                    {fld_name: "May not be before incident date/time"},
                    error_code=INVALID_ERROR,
                )



# Generated by Django 3.2.7 on 2021-11-04 11:56

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_protocol_incident", "0002_auto_20210911_2036"),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalprotocoldeviationviolation",
            name="violation",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_protocol_incident.protocolviolations",
                verbose_name="Type of violation",
            ),
        ),
        migrations.AlterField(
            model_name="protocoldeviationviolation",
            name="violation",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="edc_protocol_incident.protocolviolations",
                verbose_name="Type of violation",
            ),
        ),
    ]



from edc_action_item import ActionWithNotification
from edc_constants.constants import CLOSED, HIGH_PRIORITY

from .constants import (
    PROTOCOL_DEVIATION_VIOLATION_ACTION,
    PROTOCOL_INCIDENT_ACTION,
    WITHDRAWN,
)


class ProtocolDeviationViolationAction(ActionWithNotification):

    reference_model = None  # "my_app.protocolincident"
    admin_site_name = None  # "my_app_admin"

    name = PROTOCOL_DEVIATION_VIOLATION_ACTION
    display_name = "Submit Protocol Deviation / Violation Report"
    notification_display_name = "Protocol Deviation / Violation Report"
    parent_action_names = []
    show_link_to_changelist = True
    show_link_to_add = True
    priority = HIGH_PRIORITY

    def close_action_item_on_save(self):
        return self.reference_obj.report_status == CLOSED


class ProtocolIncidentAction(ActionWithNotification):

    reference_model = None  # "my_app.protocolincident"
    admin_site_name = None  # "my_app"

    name = PROTOCOL_INCIDENT_ACTION
    display_name = "Submit Protocol Incident Report"
    notification_display_name = "Protocol Incident Report"
    parent_action_names = []
    show_link_to_changelist = True
    show_link_to_add = True
    priority = HIGH_PRIORITY

    def close_action_item_on_save(self):
        return self.reference_obj.report_status in [CLOSED, WITHDRAWN]



from edc_constants.constants import CLOSED, OPEN

from .constants import DEVIATION, VIOLATION, WITHDRAWN

DEVIATION_VIOLATION = (
    (VIOLATION, "Protocol violation"),
    (DEVIATION, "Protocol deviation"),
)

REPORT_STATUS = (
    (OPEN, "Open. Some information is still pending."),
    (CLOSED, "Closed. This report is complete"),
    (WITHDRAWN, "Withdrawn. This report has been withdrawn after review with investigators"),
)



from typing import Tuple

from django.contrib import admin
from django.utils.html import format_html
from django_audit_fields.admin import audit_fieldset_tuple
from edc_constants.constants import CLOSED, OPEN

from edc_protocol_incident.constants import WITHDRAWN


class ProtocolIncidentModelAdminMixin:

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "subject_identifier",
                    "report_datetime",
                    "short_description",
                    "report_type",
                )
            },
        ),
        (
            "Details of protocol incident",
            {
                "fields": (
                    "safety_impact",
                    "safety_impact_details",
                    "study_outcomes_impact",
                    "study_outcomes_impact_details",
                    "incident_datetime",
                    "incident",
                    "incident_other",
                    "incident_description",
                    "incident_reason",
                )
            },
        ),
        (
            "Actions taken",
            {
                "fields": (
                    "corrective_action_datetime",
                    "corrective_action",
                    "preventative_action_datetime",
                    "preventative_action",
                    "action_required",
                )
            },
        ),
        (
            "Report status",
            {
                "fields": (
                    "report_status",
                    "reasons_withdrawn",
                    "report_closed_datetime",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    radio_fields = {
        "action_required": admin.VERTICAL,
        "report_status": admin.VERTICAL,
        "report_type": admin.VERTICAL,
        "safety_impact": admin.VERTICAL,
        "study_outcomes_impact": admin.VERTICAL,
    }

    list_filter = (
        "report_type",
        "safety_impact",
        "study_outcomes_impact",
        "report_status",
    )

    list_display = (
        "subject_identifier",
        "dashboard",
        "report_type",
        "safety_impact",
        "study_outcomes_impact",
        "report_status",
    )

    search_fields = ("subject_identifier",)

    def get_list_display(self, request) -> Tuple[str, ...]:
        list_display = super().get_list_display(request)
        custom_fields = (
            "subject_identifier",
            "dashboard",
            "description",
            "report_datetime",
            "status",
            "action_required",
            "report_type",
            "action_identifier",
            "user_created",
        )
        return custom_fields + tuple(f for f in list_display if f not in custom_fields)

    def get_list_filter(self, request) -> Tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = ("action_required", "report_status", "report_type")
        return custom_fields + tuple(f for f in list_filter if f not in custom_fields)

    def get_search_fields(self, request) -> Tuple[str, ...]:
        search_fields = super().get_search_fields(request)
        custom_fields = ("short_description",)
        return tuple(set(custom_fields + search_fields))

    def status(self, obj=None):
        color = None
        if obj.report_status == CLOSED:
            color = "green"
        elif obj.report_status == OPEN:
            color = "red"
        elif obj.report_status == WITHDRAWN:
            color = "darkorange"
        if color:
            return format_html(
                f'<span style="color:{color};">{obj.report_status.title()}</span>'
            )
        return obj.report_status.title()

    def description(self, obj=None):
        return obj.short_description.title()



from django.contrib import admin
from django_audit_fields.admin import audit_fieldset_tuple
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_model_admin.history import SimpleHistoryAdmin

from ..admin_site import edc_protocol_incident_admin
from ..forms import ProtocolDeviationViolationForm
from ..models import ProtocolDeviationViolation


@admin.register(ProtocolDeviationViolation, site=edc_protocol_incident_admin)
class ProtocolDeviationViolationAdmin(ModelAdminSubjectDashboardMixin, SimpleHistoryAdmin):

    form = ProtocolDeviationViolationForm

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "subject_identifier",
                    "report_datetime",
                    "short_description",
                    "report_type",
                )
            },
        ),
        (
            "Details of protocol violation",
            {
                "fields": (
                    "safety_impact",
                    "safety_impact_details",
                    "study_outcomes_impact",
                    "study_outcomes_impact_details",
                    "violation_datetime",
                    "violation",
                    "violation_other",
                    "violation_description",
                    "violation_reason",
                )
            },
        ),
        (
            "Actions taken",
            {
                "fields": (
                    "corrective_action_datetime",
                    "corrective_action",
                    "preventative_action_datetime",
                    "preventative_action",
                    "action_required",
                )
            },
        ),
        (
            "Report status",
            {
                "fields": (
                    "report_status",
                    "report_closed_datetime",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    radio_fields = {
        "action_required": admin.VERTICAL,
        "report_status": admin.VERTICAL,
        "report_type": admin.VERTICAL,
        "safety_impact": admin.VERTICAL,
        "study_outcomes_impact": admin.VERTICAL,
    }

    list_filter = (
        "report_type",
        "safety_impact",
        "study_outcomes_impact",
        "report_status",
    )

    list_display = (
        "subject_identifier",
        "dashboard",
        "report_type",
        "safety_impact",
        "study_outcomes_impact",
        "report_status",
    )

    search_fields = ("subject_identifier",)



from copy import deepcopy

from dateutil.relativedelta import relativedelta
from django.test.testcases import TestCase
from edc_action_item import site_action_items
from edc_constants.constants import CLOSED, NO, NOT_APPLICABLE, OPEN, OTHER
from edc_list_data import site_list_data
from edc_registration.models import RegisteredSubject
from edc_utils import get_utcnow
from edc_visit_schedule import site_visit_schedules

from edc_protocol_incident import list_data
from edc_protocol_incident.constants import DEVIATION, WITHDRAWN
from edc_protocol_incident.forms import ProtocolIncidentForm
from edc_protocol_incident.models import (
    ActionsRequired,
    ProtocolIncident,
    ProtocolViolations,
)

from ..action_items import ProtocolIncidentAction
from ..visit_schedule import visit_schedule


class TestProtocolIncident(TestCase):
    def setUp(self):
        site_action_items.registry = {}
        action_cls = ProtocolIncidentAction
        site_action_items.register(action_cls)
        site_list_data.initialize()
        site_list_data.register(list_data, app_name="edc_protocol_incident")
        site_list_data.load_data()
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)

        self.subject_identifier = "1234"
        RegisteredSubject.objects.create(subject_identifier=self.subject_identifier)

        action = ProtocolIncidentAction(subject_identifier=self.subject_identifier)
        self.data = dict(action_identifier=action.action_item.action_identifier)

    def test_incident_open_ok(self):
        data = deepcopy(self.data)
        data.update(
            {
                "subject_identifier": self.subject_identifier,
                "report_datetime": get_utcnow(),
                "report_status": OPEN,
                "reasons_withdrawn": None,
                "report_type": DEVIATION,
                "safety_impact": NOT_APPLICABLE,
                "short_description": "sdasd asd asdasd ",
                "study_outcomes_impact": NOT_APPLICABLE,
            }
        )
        obj = ProtocolIncident(**data)
        obj.save()

    def test_incident_open_form(self):
        data = deepcopy(self.data)
        data.update(
            {
                "subject_identifier": "1234",
                "report_datetime": get_utcnow(),
                "report_status": OPEN,
                "reasons_withdrawn": None,
                "report_type": DEVIATION,
                "safety_impact": NO,
                "short_description": "sdasd asd asdasd ",
                "study_outcomes_impact": NO,
                "incident_datetime": get_utcnow() - relativedelta(days=3),
                "incident": ProtocolViolations.objects.get(name=OTHER),
                "incident_other": "blah blah",
                "incident_description": "blah blah",
                "incident_reason": "blah blah",
                "violation": None,
            }
        )

        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertEqual({}, form._errors)

    def test_incident_try_to_close_form(self):
        data = deepcopy(self.data)
        data.update(
            {
                "subject_identifier": "1234",
                "report_datetime": get_utcnow() - relativedelta(days=1),
                "report_status": OPEN,
                "reasons_withdrawn": None,
                "report_type": DEVIATION,
                "safety_impact": NO,
                "short_description": "sdasd asd asdasd ",
                "study_outcomes_impact": NO,
                "incident_datetime": get_utcnow() - relativedelta(days=3),
                "incident": ProtocolViolations.objects.get(name=OTHER),
                "incident_other": "blah blah",
                "incident_description": "blah blah",
                "incident_reason": "blah blah",
                "violation": None,
            }
        )

        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        # self.assertIn("corrective_action_datetime", form._errors)
        data.update(corrective_action_datetime=get_utcnow())
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertIn("corrective_action", form._errors)

        data.update(corrective_action="we took corrective action")
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        # self.assertIn("preventative_action_datetime", form._errors)

        data.update(preventative_action_datetime=get_utcnow())
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertIn("preventative_action", form._errors)

        data.update(preventative_action="we took preventative action", report_status=CLOSED)
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertIn("action_required", form._errors)

        data.update(action_required=ActionsRequired.objects.get(name="remain_on_study"))
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertIn("report_closed_datetime", form._errors)

        data.update(report_closed_datetime=get_utcnow())
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertEqual({}, form._errors)

        data.update(report_status=WITHDRAWN)
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertIn("reasons_withdrawn", form._errors)

        data.update(reasons_withdrawn="cause i feel like it")
        form = ProtocolIncidentForm(data=data, instance=ProtocolIncident())
        form.is_valid()
        self.assertEqual({}, form._errors)
